
/*! yuzu - v2.0.0-rc.16
 * https://github.com/dwightjack/yuzu/packages/yuzu
 * Copyright (c) 2019 - Marco Solazzi;
 * Licensed MIT
 */

import { warn, qs, isElement, bindMethod, nextUid, qsa, evaluate, isPlainObject, Events, noop } from 'yuzu-utils';

/*! *****************************************************************************
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License"); you may not use
this file except in compliance with the License. You may obtain a copy of the
License at http://www.apache.org/licenses/LICENSE-2.0

THIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED
WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
MERCHANTABLITY OR NON-INFRINGEMENT.

See the Apache Version 2.0 License for specific language governing permissions
and limitations under the License.
***************************************************************************** */
/* global Reflect, Promise */

var extendStatics = function(d, b) {
    extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return extendStatics(d, b);
};

function __extends(d, b) {
    extendStatics(d, b);
    function __() { this.constructor = d; }
    d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
}

function __rest(s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) if (e.indexOf(p[i]) < 0)
            t[p[i]] = s[p[i]];
    return t;
}

function __awaiter(thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
}

function __generator(thisArg, body) {
    var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
    function verb(n) { return function (v) { return step([n, v]); }; }
    function step(op) {
        if (f) throw new TypeError("Generator is already executing.");
        while (_) try {
            if (f = 1, y && (t = op[0] & 2 ? y["return"] : op[0] ? y["throw"] || ((t = y["return"]) && t.call(y), 0) : y.next) && !(t = t.call(y, op[1])).done) return t;
            if (y = 0, t) op = [op[0] & 2, t.value];
            switch (op[0]) {
                case 0: case 1: t = op; break;
                case 4: _.label++; return { value: op[1], done: false };
                case 5: _.label++; y = op[1]; op = [0]; continue;
                case 7: op = _.ops.pop(); _.trys.pop(); continue;
                default:
                    if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                    if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                    if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                    if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                    if (t[2]) _.ops.pop();
                    _.trys.pop(); continue;
            }
            op = body.call(thisArg, _);
        } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
        if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
    }
}

function __read(o, n) {
    var m = typeof Symbol === "function" && o[Symbol.iterator];
    if (!m) return o;
    var i = m.call(o), r, ar = [], e;
    try {
        while ((n === void 0 || n-- > 0) && !(r = i.next()).done) ar.push(r.value);
    }
    catch (error) { e = { error: error }; }
    finally {
        try {
            if (r && !r.done && (m = i["return"])) m.call(i);
        }
        finally { if (e) throw e.error; }
    }
    return ar;
}

function __spread() {
    for (var ar = [], i = 0; i < arguments.length; i++)
        ar = ar.concat(__read(arguments[i]));
    return ar;
}

var LISTENER_REGEXP = /^([^ ]+)(?: (.+))?$/;
var objDiff = noop;
if (process.env.NODE_ENV !== 'production') {
    objDiff = function objDiff(match, obj, msg) {
        var keys = Object.keys(match);
        var keyStr = keys.length > 0 ? keys.join(', ') : 'no keys';
        Object.keys(obj).forEach(function (k) {
            if (keys.indexOf(k) === -1) {
                msg(k, keyStr);
            }
        });
    };
}
/**
 * `Component` is an extensible class constructor which provides the building block of Yuzu component system.
 *
 * It inherits methods from [**Events**](/packages/utils/api/events).
 *
 * ```js
 * const instance = new Component({ ... })
 * ```
 *
 * > **Lifecycle**
 * >
 * > | stage    | hooks     |
 * > |----------| --------- |
 * > | `create` | `created` |
 *
 * @class
 * @param {object} [options={}] Instance options
 * @property {boolean} $active `true` if the instance is mounted and initialized
 * @property {object} options Instance options (see [defaultOptions](#defaultOptions))
 * @property {object} state Instance state (see [setState](#setState))
 * @property {Element} $el The instance root DOM element (see [mount](#mount))
 * @property {Object.<string, Element>} $els  Object mapping references to component's child DOM elements (see `selectors` below)
 * @property {{ string: Component }} $refs Object mapping references to child components (see [setRef](#setRef))
 * @property {object} selectors Object mapping a child element's reference name and a CSS selector or custom function
 * @property {Object.<string, function|string>} listeners Object mapping DOM listeners and handlers (see [setListener](#setListener))
 * @property {Object.<string, function|string>} actions Object mapping state keys and functions to executed on state update
 * @returns {Component}
 */
var Component = /** @class */ (function (_super) {
    __extends(Component, _super);
    // public $$getTree?: IObject;
    /**
     * ```js
     * this.readyState(state, prevState)
     * ```
     *
     * Optional method used to delay the execution of [`ready`](#ready) after the state satisfies to a given condition.
     *
     * Will be executed at every state change until it returns `true`.
     *
     * @name readyState
     * @public
     * @type {function}
     * @param {object} state The current state
     * @param {object} prevState The previous state
     * @returns {boolean}
     */
    /**
     * Component constructor
     */
    function Component(options) {
        if (options === void 0) { options = {}; }
        var _this = _super.call(this) || this;
        _this.$warn = warn(_this);
        var defaultOptionsFn = _this.constructor
            .defaultOptions;
        var defaultOptions;
        if (typeof defaultOptionsFn === 'function') {
            _this.$warn('the static property `defaultOptions` is deprecated. Please move the method to the instance');
            defaultOptions = defaultOptionsFn(_this);
        }
        else {
            defaultOptions = _this.defaultOptions(_this) || {};
        }
        _this.state = {};
        if (process.env.NODE_ENV !== 'production') {
            objDiff(defaultOptions, options, function (k, keys) {
                return _this.$warn("Option " + k + "\" has been discarded because it is not defined in component's defaultOptions. Accepted keys are: " + keys);
            });
        }
        var entries = Object.entries(defaultOptions);
        _this.options = entries.reduce(function (opts, _a) {
            var _b = __read(_a, 2), key = _b[0], value = _b[1];
            var v = options[key] !== undefined ? options[key] : value;
            if (typeof v === 'function' && !Component.isComponent(v)) {
                v = v.bind(_this);
            }
            opts[key] = v; // eslint-disable-line no-param-reassign
            return opts;
        }, {});
        _this.$active = false;
        // DOM references
        _this.$els = {};
        // sub components references
        _this.$refs = {};
        _this.$refsStore = new Map();
        _this.$listeners = new Map();
        _this.created();
        return _this;
    }
    /**
     * ```js
     * Component.isComponent(obj)
     * ```
     *
     * Checks whether the passed-in value is a Component constructor.
     *
     * @static
     * @param {*} value
     * @returns {boolean}
     */
    Component.isComponent = function (value) {
        return !!(value && value.YUZU_COMPONENT);
    };
    Component.prototype.defaultOptions = function () {
        return {};
    };
    /**
     * ```js
     * mount(el, [state])
     * ```
     *
     * Mounts a component instance on a DOM element and initializes it.
     *
     * ?> To prevent initialization and just mount the component set `state` to `null`.
     *
     * > **Lifecycle**
     * >
     * > | stage    | hooks     |
     * > |----------| --------- |
     * > | `mount` | `beforeMount` <sup>(1)</sup>, `mounted` |
     *
     * 1. Just after attaching the root element (`this.$el`) but before any listener and selector registration.
     *
     * @param {string|Element} el Component's root element
     * @param {object|null} [state={}] Initial state
     * @returns {Component}
     */
    Component.prototype.mount = function (el, state) {
        var _this = this;
        if (state === void 0) { state = {}; }
        if (this.$el) {
            throw new Error('Component is already mounted');
        }
        if (this.detached) {
            throw new Error('You cannot mount a detached component. Please use `init` instead');
        }
        var $el = typeof el === 'string' ? qs(el) : el;
        if (!isElement($el)) {
            // fail silently (kinda...);
            this.$warn('Element is not a DOM element', $el);
            return this;
        }
        this.$el = $el; // eslint-disable-line no-multi-assign
        this.beforeMount();
        if (this.selectors) {
            Object.entries(this.selectors).forEach(function (_a) {
                var _b = __read(_a, 2), key = _b[0], selector = _b[1];
                if (typeof selector === 'function') {
                    _this.$els[key.replace('[]', '')] = selector(_this.$el, _this.options);
                    return;
                }
                if (!key.endsWith('[]')) {
                    var el_1 = _this.findNode(selector);
                    if (el_1) {
                        _this.$els[key] = el_1;
                    }
                }
                else {
                    _this.$els[key.slice(0, -2)] = _this.findNodes(selector);
                }
            });
        }
        if (this.listeners) {
            Object.entries(this.listeners).forEach(function (_a) {
                var _b = __read(_a, 2), def = _b[0], handler = _b[1];
                _this.setListener(def, bindMethod(_this, handler));
            });
        }
        if (state) {
            this.init(state);
        }
        this.mounted();
        return this;
    };
    /**
     * ```js
     * init([state])
     * ```
     * Initializes the component instance.
     *
     * > **Lifecycle**
     * >
     * > | stage    | hooks     |
     * > |----------| --------- |
     * > | `init` | `initialize`, `ready` |
     *
     * @param {object|null} [state={}] Initial state
     * @returns {Component}
     */
    Component.prototype.init = function (state) {
        var _this = this;
        if (state === void 0) { state = {}; }
        if (!this.detached && !isElement(this.$el)) {
            throw new Error('component instance not mounted');
        }
        var $el = this.$el;
        // initialization placeholder
        var uid = $el && $el.getAttribute(Component.UID_DATA_ATTR);
        if (uid) {
            this.$warn("Element " + uid + " is already initialized... skipping", $el);
            this.$uid = uid;
            return this;
        }
        uid = nextUid();
        this.$uid = uid;
        if ($el) {
            $el.setAttribute(Component.UID_DATA_ATTR, uid);
            if (!$el.id) {
                $el.id = "c_" + uid;
            }
        }
        this.initialize();
        if (this.actions) {
            Object.entries(this.actions).forEach(function (_a) {
                var _b = __read(_a, 2), key = _b[0], method = _b[1];
                _this.on("change:" + key, bindMethod(_this, method));
            });
        }
        var initialState = Object.assign(this.state, state);
        this.replaceState(initialState);
        this.$active = true;
        if (this.readyState) {
            // is it a promise ?
            var watcher_1 = function (current, prev) {
                if (_this.readyState(current, prev)) {
                    _this.off('change:*', watcher_1);
                    _this.ready();
                }
            };
            this.on('change:*', watcher_1);
            return this;
        }
        this.ready();
        return this;
    };
    /**
     * Lifecycle hook called on instance creation.
     *
     * At this stage just the instance options (`this.options`) are initialized.
     *
     *  Overwrite this method with custom logic in your components.
     *
     * ?> Use this hook to tap as early as possible into the component's properties. For example to set a dynamic initial state.
     */
    Component.prototype.created = function () { };
    /**
     * Lifecycle hook called just before mounting the instance onto the root element.
     *
     * At this stage the component is already attached to its root DOM element.
     *
     * Overwrite this method with custom logic in your components.
     */
    Component.prototype.beforeMount = function () { };
    /**
     * Lifecycle hook called when the instance gets mounted onto a DOM element.
     *
     * At this stage both children elements (`this.$els.*`) and event listeners configured in `this.listeners` have been setup.
     *
     * Overwrite this method with custom logic in your components.
     *
     * ?> Use this method when you need to work with the DOM or manage any side-effect that requires the component to be into the DOM.
     */
    Component.prototype.mounted = function () { };
    /**
     * Lifecycle hook called before instance initialization.
     *
     * At this stage the state and state listeners are not yet been initialized.
     *
     * Overwrite this method with custom logic in your components.
     *
     * ?> Use this method to set child components by [setRef](#setRef) and run any preparatory work on the instance.
     */
    Component.prototype.initialize = function () { };
    /**
     * Lifecycle hook called after instance initialization.
     *
     * At this stage State and event binding are already in place.
     *
     * Overwrite this method with custom logic in your components.
     *
     * ?> `ready` lifecycle can be delayed (_async ready_) by implementing a [`readyState`](packages/yuzu/#async-ready-state) method.
     */
    Component.prototype.ready = function () { };
    /**
     * Lifecycle hook called just before closing child refs.
     *
     * This hook is called just before destroying the instance. Every property, listener and state feature is still active.
     *
     * Overwrite this method with custom logic in your components.
     *
     * !> This is an async method. Return a promise in order to suspend the destroy process.
     */
    Component.prototype.beforeDestroy = function () { };
    /**
     * Returns an array of elements matching a CSS selector in the context of the component's root element
     *
     * @param selector {string} CSS selector to match
     * @return {Element[]}
     */
    Component.prototype.findNodes = function (selector) {
        return qsa(selector, this.$el);
    };
    /**
     * Returns the first element matching a CSS selector in the context of the component's root element
     *
     * @param selector {string} CSS selector to match
     * @return {Element}
     */
    Component.prototype.findNode = function (selector) {
        return qs(selector, this.$el);
    };
    /**
     * ```js
     * getState(key)
     * ```
     *
     * Returns a property of the state or a default value if the property is not set.
     *
     * ?> In ES6 environments you can use a [destructuring assignment](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Operators/Destructuring_assignment) instead: ` const { name = 'John'} = this.state`
     *
     *
     * @param {string} key Property to return
     * @param {*} def Default value
     * @returns {*}
     * @example
     * const instance = new Component().mount('#app', { a: 1 });
     * // instance.getState('a') === 1
     *
     * // instance.getState('b', false) === false
     */
    Component.prototype.getState = function (key, def) {
        return this.state.hasOwnProperty(key) ? this.state[key] : def;
    };
    /* eslint-disable class-methods-use-this */
    /**
     * ```js
     * shouldUpdateState(string, currentValue, newValue)
     * ```
     * Executes a strict inequality comparison (`!==`) on the passed-in values and returns the result.
     *
     * !> This method is executed on every [`setState`](#setstate) call.
     *
     * You can overwrite this method with your own validation logic.
     *
     * @param {string} key State property name
     * @param {*} currentValue Value stored in the current state
     * @param {*} newValue New value
     * @returns {boolean}
     */
    Component.prototype.shouldUpdateState = function (key, currentValue, newValue) {
        return currentValue !== newValue;
    };
    /* eslint-enable class-methods-use-this */
    /**
     * ```js
     * setState(updater, [silent])
     * ```
     *
     * Updates the internal instance state by creating a shallow copy of the current state and updating the passed-in keys.
     *
     * If the computed new state is different from the old one it emits a `change:<property>` event for every changed property
     * as well as a global `change:*` event.
     *
     * If the new value argument is a function, it will be executed with the current state as argument.
     * The returned value will be used to update the state.
     *
     * ?> To prevent an instance from emitting `change:` events set the second argument to `true` (silent update).
     *
     * @param {object|function} updater Defines which part of the state must be updated
     * @param {boolean} [silent=false] Update the state without emitting change events
     * @example
     * instance.on('change:a', (next, prev) => console.log(next, prev));
     * instance.setState({ a: 1 }); //emits 'change:a' -> logs undefined,1
     *
     * instance.setState({ a: 1 }); //nothing happens
     * instance.setState({ a: 2 }, true); //nothing happens, again...
     *
     * // use the current state to calculate its next value
     * instance.setState(({ a }) => ({a + 1}));
     */
    Component.prototype.setState = function (updater, silent) {
        var _this = this;
        if (silent === void 0) { silent = false; }
        var changed = [];
        var prevState = this.state;
        var changeSet = evaluate(updater, this.state);
        if (process.env.NODE_ENV !== 'production') {
            objDiff(this.state, changeSet, function (k, keys) {
                return "setState: key \"" + k + "\" has been discarded because it is not defined in the component's initial state. Accepted keys are: " + keys;
            });
        }
        var entries = Object.entries(this.state);
        this.state = entries.reduce(function (newState, _a) {
            var _b = __read(_a, 2), k = _b[0], prevValue = _b[1];
            var value = changeSet[k];
            if (value === undefined ||
                _this.shouldUpdateState(k, prevValue, value) === false) {
                newState[k] = prevValue; // eslint-disable-line no-param-reassign
            }
            else {
                changed.push(k);
                newState[k] = value; // eslint-disable-line no-param-reassign
            }
            return newState;
        }, {});
        if (!silent && changed.length > 0) {
            while (changed.length > 0) {
                var k = changed.pop();
                this.emit("change:" + k, this.state[k], prevState[k]);
            }
            this.emit('change:*', this.state, prevState);
        }
    };
    /**
     * ```js
     * replaceState(newState, [silent])
     * ```
     *
     * Replaces the current state of the instance with a completely new state.
     *
     * !> Note that this methods is un-affected by [`shouldUpdateState`](#shouldupdatestate).
     *
     * @param {object} newState The new state object
     * @param {boolean} [silent=false] Replace the state without emitting change events
     * @example
     * instance.replaceState({ a: 1 });
     * // instance.state.a === 1
     * instance.replaceState({ b: 2 });
     * // instance.state.b === 2
     * // instance.state.a === undefined
     */
    Component.prototype.replaceState = function (newState, silent) {
        var _this = this;
        if (silent === void 0) { silent = false; }
        var prevState = this.state;
        this.state = Object.assign({}, newState);
        var entries = Object.entries(this.state);
        entries.forEach(function (_a) {
            var _b = __read(_a, 2), key = _b[0], value = _b[1];
            if (!silent) {
                _this.emit("change:" + key, value, prevState[key]);
            }
        });
        if (!silent) {
            this.emit('change:*', this.state, prevState);
        }
    };
    /**
     * ```js
     * broadcast(event, [...params])
     * ```
     *
     * Emits a `broadcast:<eventname>` event on every child component listed in `$refs`.
     *
     * @param {string} event Event name
     * @param {*[]} [params] Additional arguments to pass to the handler
     * @example
     * const child = new Component('#child');
     * child.on('broadcast:log', (str) => console.log(str));
     *
     * instance.setRef({ id: 'child', component: child });
     * instance.broadcast('log', 'test') // child component logs 'test'
     */
    Component.prototype.broadcast = function (event) {
        var params = [];
        for (var _i = 1; _i < arguments.length; _i++) {
            params[_i - 1] = arguments[_i];
        }
        this.$refsStore.forEach(function (instance) {
            instance.emit.apply(instance, __spread(["broadcast:" + event], params));
        });
    };
    /**
     * ```js
     * setListener(string, handler)
     * ```
     *
     * Sets a DOM event listener.
     *
     * The first argument must be a string composed by an event name (ie `click`) and a CSS selector (`.element`)
     * separated by a space.
     *
     * If the CSS selector starts with `@` the listener will be attached to the
     * corresponding reference child element (`this.$els.<element>`), if any.
     *
     * @param {string} def Event and target element definition. Format `eventName [target]`
     * @param {function} handler Event handler
     * @example
     *
     * // attach a click handler to a child element
     * instance.setListener('click .button', () => ...)
     *
     * // attach a click handler to this.$els.btn
     * instance.setListener('click @btn', () => ...)
     *
     * // attach a click handler to this.$el
     * instance.setListener('click', () => ...)
     */
    Component.prototype.setListener = function (def, handler) {
        var _this = this;
        var _a, _b;
        var event;
        var selector;
        var match = def && def.match(LISTENER_REGEXP);
        if (match) {
            var $el = this.$el;
            _a = __read(match, 3), event = _a[1], _b = _a[2], selector = _b === void 0 ? $el : _b;
            var element = void 0;
            if (typeof selector === 'string') {
                element = selector.startsWith('@')
                    ? this.$els[selector.slice(1)]
                    : qs(selector, $el);
            }
            else {
                element = selector;
            }
            if (element) {
                if (Array.isArray(element)) {
                    element.forEach(function (el, i) {
                        var h = function (e) {
                            handler.call(_this, e, i);
                        };
                        el.addEventListener(event, h);
                        _this.$listeners.set(h, { event: event, element: el });
                    });
                }
                else {
                    element.addEventListener(event, handler);
                    this.$listeners.set(handler, { event: event, element: element });
                }
            }
        }
    };
    /**
     * ```js
     * removeListeners()
     * ```
     *
     * Removes all DOM event listeners attached with `.setListener`.
     *
     */
    Component.prototype.removeListeners = function () {
        this.$listeners.forEach(function (_a, handler) {
            var event = _a.event, element = _a.element;
            element.removeEventListener(event, handler);
        });
        this.$listeners.clear();
    };
    /**
     * ```js
     * setRef(config, [props])
     * ```
     *
     * Attaches a reference to a child component.
     *
     * If a reference `id` is already attached, the previous one is destroyed and replaced with the new one.
     *
     * ?> This is an async method returning a promise.
     *
     * @param {object} config A child component configuration object
     * @param {string} config.id Reference id. Will be used to set a reference to the child component onto `this.$refs`
     * @param {component} config.component Component constructor or component instance
     * @param {string|HTMLElement} [config.el] Child component root element. This property is ignored if `config.component` is a component instance or a detached component constructor
     * @param {Object} [config.on] Child component event listeners. Format `{ 'eventname': handler }`
     * @param {*} config.* Any other property listed here will be passed to the constructor as option
     * @param {object} [props] Child component initial state
     * @returns {Promise}
     * @example
     * const parent = new Component('#root');
     *
     * class ChildComponent extends Component {}
     *
     * // as constructor
     * parent.setRef({
     *   id: 'child',
     *   component: ChildComponent,
     *   el: '#child',
     *   // other options here...
     * });
     *
     * // as instance
     * parent.setRef({
     *   id: 'child',
     *   component: new ChildComponent({ ... }).mount('#child', null) // <-- prevent component init
     * });
     *
     * // sync parent - child state
     * // updates child `parentCount` state whenever `parent.state.count` changes
     * parent.setRef({
     *   id: 'child',
     *   component: ChildComponent,
     *   el: '#child',
     *   // other options here...
     * }, {
     *   parentCount: (parentState) => parentState.count
     * });
     */
    Component.prototype.setRef = function (refCfg, props) {
        return __awaiter(this, void 0, void 0, function () {
            var ChildComponent, el, id, on, options, detached, ref, $refs, prevRef, refState, stateMap, $el, $parent;
            var _this = this;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        if (!isPlainObject(refCfg)) {
                            throw new TypeError('Invalid reference configuration');
                        }
                        ChildComponent = refCfg.component, el = refCfg.el, id = refCfg.id, on = refCfg.on, options = __rest(refCfg, ["component", "el", "id", "on"]);
                        detached = this.detached;
                        // if (el && detached) {
                        //   throw new Error(
                        //     `setRef "${id}": you cannot define a component with DOM root as child of a detached component.`,
                        //   );
                        // }
                        if (Component.isComponent(ChildComponent)) {
                            ref = new ChildComponent(options);
                        }
                        else if (ChildComponent instanceof Component) {
                            ref = ChildComponent;
                        }
                        else if (typeof ChildComponent === 'function') {
                            ref = ChildComponent(this.$el, this.state);
                        }
                        else {
                            throw new TypeError('Invalid reference configuration');
                        }
                        // $context gets propagated to every child component
                        // this way we can use `connect`-ed components as reference
                        if (this.$context) {
                            Object.defineProperty(ref, '$context', {
                                enumerable: false,
                                get: function () { return _this.$context; },
                            });
                        }
                        Object.defineProperty(ref, '$parent', {
                            writable: true,
                            enumerable: false,
                            value: this,
                        });
                        if (!id) {
                            throw new Error('Invalid reference id string');
                        }
                        // bind events...
                        if (on) {
                            Object.entries(on).forEach(function (_a) {
                                var _b = __read(_a, 2), name = _b[0], handler = _b[1];
                                ref.on(name, handler);
                            });
                        }
                        $refs = this.$refs;
                        prevRef = $refs[id];
                        refState = {};
                        $refs[id] = ref;
                        this.$refsStore.set(id, ref);
                        if (!ref.detached && !ref.$el) {
                            if (!el) {
                                throw new Error("You need to provide a root element for the child element with id \"" + id + "\".");
                            }
                            ref.mount(el, null);
                        }
                        stateMap = evaluate(props, ref, this);
                        if (stateMap) {
                            Object.entries(stateMap).forEach(function (_a) {
                                var _b, _c;
                                var _d = __read(_a, 2), name = _d[0], value = _d[1];
                                if (typeof value === 'function') {
                                    var key_1 = name;
                                    var src = '*';
                                    if (name.includes('>')) {
                                        _b = __read(name.split('>'), 2), _c = _b[0], src = _c === void 0 ? '*' : _c, key_1 = _b[1];
                                    }
                                    refState[key_1] = value(src !== '*' ? _this.state[src] : _this.state, ref);
                                    _this.on("change:" + src, function (state) {
                                        var _a;
                                        ref.setState((_a = {}, _a[key_1] = value(state, ref), _a));
                                    });
                                }
                                else {
                                    refState[name] = value;
                                }
                            });
                        }
                        $el = this.$el;
                        if (detached && ref.$el) {
                            $parent = this.$parent;
                            while (!$el && $parent) {
                                $el = $parent.$el;
                                $parent = $parent.$parent;
                            }
                        }
                        if (!$el && ref.$el) {
                            throw new Error("You cannot attach a plain Component to a DetachedComponents tree. (" + id + ")");
                        }
                        if (!prevRef) return [3 /*break*/, 2];
                        return [4 /*yield*/, prevRef.destroy()];
                    case 1:
                        _a.sent();
                        if (prevRef.$el &&
                            $el.contains(prevRef.$el) &&
                            prevRef.$el.parentElement &&
                            ref.$el) {
                            prevRef.$el.parentElement.replaceChild(ref.$el, prevRef.$el);
                        }
                        else if (ref.$el && !$el.contains(ref.$el)) {
                            $el.appendChild(ref.$el);
                        }
                        return [2 /*return*/, ref.init(refState)];
                    case 2:
                        if ($el && ref.$el && !$el.contains(ref.$el)) {
                            $el.appendChild(ref.$el);
                        }
                        return [2 /*return*/, ref.init(refState)];
                }
            });
        });
    };
    /**
     * Destroys and detaches a specific child component by its reference `id` (as set in `setRef`).
     *
     * @param {string} id Child component reference id
     * @param {boolean} [detach=false] Remove the child component root element from the DOM
     * @returns {Promise}
     */
    Component.prototype.destroyRef = function (id, detach) {
        if (detach === void 0) { detach = false; }
        return __awaiter(this, void 0, void 0, function () {
            var $refsStore, ref;
            return __generator(this, function (_a) {
                $refsStore = this.$refsStore;
                ref = $refsStore.get(id);
                if (!ref) {
                    throw new Error("Child component \"" + id + "\" not found.");
                }
                $refsStore["delete"](id);
                delete this.$refs[id];
                if (!detach) {
                    return [2 /*return*/, ref.destroy()];
                }
                return [2 /*return*/, ref.destroy().then(function () {
                        if (ref.$el) {
                            ref.$el.remove();
                        }
                    })];
            });
        });
    };
    /**
     * Calls `.destroy()` on every child references and detaches them from the parent component.
     *
     * !> This is an async method returning a promise
     *
     * @returns {Promise}
     */
    Component.prototype.destroyRefs = function () {
        return __awaiter(this, void 0, void 0, function () {
            var $refsStore, result, e_1;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        $refsStore = this.$refsStore;
                        if ($refsStore.size === 0) {
                            // exit early!
                            return [2 /*return*/, Promise.resolve()];
                        }
                        _a.label = 1;
                    case 1:
                        _a.trys.push([1, 3, , 4]);
                        return [4 /*yield*/, Promise.all(__spread($refsStore.values()).map(function (ref) { return ref.destroy(); }))];
                    case 2:
                        result = _a.sent();
                        this.$refs = {};
                        $refsStore.clear();
                        return [2 /*return*/, result];
                    case 3:
                        e_1 = _a.sent();
                        this.$warn('An error occurred while destroy the component child components', e_1);
                        return [2 /*return*/, Promise.reject(e_1)];
                    case 4: return [2 /*return*/];
                }
            });
        });
    };
    /**
     * ```js
     * destroy()
     * ```
     *
     * Detaches DOM events, instance's events and destroys all references as well.
     *
     * > **Lifecycle**
     * >
     * > | stage    | hooks     |
     * > |----------| --------- |
     * > | `destroy` | `beforeDestroy` |
     *
     *
     * !> This is an async method returning a promise
     *
     * @returns {Promise}
     */
    Component.prototype.destroy = function () {
        return __awaiter(this, void 0, void 0, function () {
            var e_2;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4 /*yield*/, this.beforeDestroy()];
                    case 1:
                        _a.sent();
                        this.removeListeners();
                        if (this.$parent) {
                            // clear the reference to the parent component
                            this.$parent = undefined;
                        }
                        this.off();
                        if (this.$el) {
                            this.$el.removeAttribute(Component.UID_DATA_ATTR); // eslint-disable-line no-console
                        }
                        _a.label = 2;
                    case 2:
                        _a.trys.push([2, 4, , 5]);
                        return [4 /*yield*/, this.destroyRefs()];
                    case 3:
                        _a.sent();
                        this.$active = false;
                        return [2 /*return*/, Promise.resolve()];
                    case 4:
                        e_2 = _a.sent();
                        this.$warn('destroy error: ', e_2);
                        return [2 /*return*/, Promise.reject(e_2)];
                    case 5: return [2 /*return*/];
                }
            });
        });
    };
    /**
     * ```js
     * Component.UID_DATA_ATTR
     * ```
     * Component root element attribute marker.
     *
     * @static
     * @returns {object}
     */
    Component.UID_DATA_ATTR = 'data-cid';
    /**
     * Marks yuzu components
     *
     * @static
     */
    Component.YUZU_COMPONENT = true;
    return Component;
}(Events));

/**
 * `DetachedComponent` is a special kind of `Component` that does not have a root element (`$el`).
 *
 * @class
 * @see [Component](/packages/yuzu/api/component)
 * @returns {DetachedComponent}
 */
var DetachedComponent = /** @class */ (function (_super) {
    __extends(DetachedComponent, _super);
    function DetachedComponent() {
        var _this = _super !== null && _super.apply(this, arguments) || this;
        _this.detached = true;
        return _this;
    }
    return DetachedComponent;
}(Component));

/**
 * Element array Iterator.
 *
 * The first argument can be:
 *
 * - a CSS selector. Element will be matched in the context of the passed-in parent's [`$el`](/packages/yuzu/api/component#instance-properties) property.
 * - an array of elements
 * - a function receiving the parent's root element and returning an array of element.
 *
 * The second parameter is an iterator function receiving the current element and its index as arguments.
 *
 * Returns a function that accepts a parent component as first argument and iterates the iterator function over an array of DOM elements.
 *
 * @param {string|Element[]} selector Elements CSS selector, array of elements or a function returning an array of elements.
 * @param {function} fn Iterator function
 * @example
 * import { Children } from 'yuzu';
 *
 * const parent = new ParentComponent().mount('#list');
 * const iterator = (el, i) => new ChildComponent(el, { index: i });
 *
 * const childComponentArray = Children('.items', iterator)(parent);
 *
 * // same as
 * // Children((el) => Array.from(el.querySelectorAll('.items')), iterator)(parent);
 *
 * // same as
 * // const els = Array.from(el.querySelectorAll('.items'))
 * // Children(els , iterator)(parent);
 */
var Children = function (selector, fn) { return function (ctx) {
    var els;
    if (typeof selector === 'string') {
        els = qsa(selector, ctx.$el);
    }
    else if (typeof selector === 'function') {
        els = selector(ctx.$el);
    }
    else if (Array.isArray(selector)) {
        els = selector;
    }
    else {
        throw new TypeError('Invalid selector argument');
    }
    return els.map(fn);
}; };

var childRefIdx = 0;
/**
 * `mount` is an helper function to setup trees of components in a functional way.
 *
 * Returns a mount function which in turn accepts a parent component. If present, the returning instance will be attached as reference to the parent component.
 *
 * Child components listed in the `children` parameter will be automatically set as references in the component (uses: [`Component#setRef`](/packages/yuzu/api/component#setref))
 *
 * See the guide on **[functional composition](/packages/yuzu/#functional-composition)** for implementation examples.
 *
 * @param {Component} ComponentConstructor A component constructor (either created by extending `Component` or by [`extend`](/packages/yuzu/api/extend))
 * @param {HTMLElement|string|null} el A mount DOM node (either a CSS selector string or a DOM element). Can be `null` if the component is a [detached component](/packages/yuzu/#detached-components)
 * @param {function[]|function} [children] Child components. Either an array of `mount` functions or a function returning an array of mount functions
 * @param {object} [props] Mount props
 * @param {string} [props.id] Optional component id (used to create a reference onto the parent component)
 * @param {object} [props.state] Component initial state
 * @param {*} [props.*] every other property will be passed as instance option
 * @return {function}
 */
function mount(ComponentConstructor, el, props, children) {
    if (props === void 0) { props = {}; }
    var _a = props || {}, _b = _a.state, state = _b === void 0 ? {} : _b, id = _a.id, _c = _a.on, on = _c === void 0 ? {} : _c, options = __rest(_a, ["state", "id", "on"]);
    var component = new ComponentConstructor(options);
    return function mounter(ctx) {
        if (!component.detached) {
            var root = typeof el === 'string' ? qs(el, ctx && ctx.$el) : el;
            if (isElement(root)) {
                component.mount(root, ctx ? null : state);
            }
        }
        else if (component.detached && !ctx) {
            component.init(state);
        }
        if (ctx) {
            ctx.setRef({
                component: component,
                id: id || "ref__" + ++childRefIdx,
                on: evaluate(on, ctx),
            }, state);
        }
        var childrenList = Array.isArray(children) ? children : [];
        if (typeof children === 'function') {
            childrenList.push.apply(childrenList, __spread(children(component)));
        }
        for (var i = 0, l = childrenList.length; i < l; i += 1) {
            var child = childrenList[i];
            child(component);
        }
        return component;
    };
}

/**
 * Utility method to create new Component classes in environments that don't support ES6 `class`es.
 *
 * @param {Component} parent Component class to extend
 * @param {object} props New component default properties and methods
 * @example
 * // UMD environment
 * const Text = YZ.extend(YZ.Component, {
 *  created: function () {
 *    this.state = { body:  '' };
 *  },
 *  mounted: function () {
 *    this.$el.innerHTML = this.state.body;
 *  }
 * });
 */
var extend = function (parent, props) {
    if (props === void 0) { props = {}; }
    var child = props.hasOwnProperty('constructor')
        ? props.constructor
        : function ChildConstructor(options) {
            return parent.call(this, options);
        };
    // https://github.com/mridgway/hoist-non-react-statics/blob/master/index.js#L51
    var keys = Object.getOwnPropertyNames(parent);
    for (var i = 0, l = keys.length; i < l; i += 1) {
        var key = keys[i];
        var descriptor = Object.getOwnPropertyDescriptor(parent, key);
        try {
            // Avoid failures from read-only properties
            Object.defineProperty(child, key, descriptor);
        }
        catch (e) { }
    }
    child.prototype = Object.assign(Object.create(parent.prototype), props);
    child.prototype.constructor = child;
    child.__super__ = parent.prototype;
    return child;
};

/**
 * `devtools` is an helper function that will expose the instance of a Component in a `$yuzu` property attached to its root DOM element.
 *
 * It will also extend every Component instance with some useful methods.
 *
 * !> To improve performance in production, this property and its methods will be available just when `process.env.NODE_ENV !== 'production'`.
 *
 * To initialize the devtools copy the following code into your entry point:
 *
 * ```js
 * import { Component, devtools } from 'yuzu';
 *
 * devtools(Component);
 * ```
 */
var devtools;
if (process.env.NODE_ENV !== 'production') {
    /**
     *
     * @private
     * @param {string} label Logger name
     */
    var createStateLogger_1 = function (label) {
        var $listeners = new Map();
        /**
         * @namespace logger
         * @private
         */
        return {
            label: label,
            /**
             * Subscribes to a specific event and logs its arguments when emitted. Returns an unsubscribe function.
             *
             * @private
             * @see logger#log
             * @param {Component} instance Target instance
             * @param {string} [event='change:*'] Event to subscribe to
             * @returns {function}
             */
            subscribe: function (instance, event) {
                var _this = this;
                if (event === void 0) { event = 'change:*'; }
                var key = event === 'change:*' ? 'change' : event;
                var listener = this.log.bind(this, key);
                if ($listeners.has(event)) {
                    console.warn("[yuzu-devtools:" + this.label + "]Already listening for \"" + event + "\" on " + instance.$uid);
                }
                instance.on(event, listener);
                $listeners.set(event, listener);
                return function () {
                    _this.unsubscribe(instance, event);
                };
            },
            /**
             * Unsubscribes the logger from a specific event.
             *
             * @private
             * @param {Component} instance Target instance
             * @param {string} [event='change:*'] Event to subscribe to
             */
            unsubscribe: function (instance, event) {
                if (event === void 0) { event = 'change:*'; }
                if ($listeners.has(event)) {
                    instance.off(event, $listeners.get(event));
                    $listeners["delete"](event);
                }
            },
            /**
             * Unsubscribes the logger from every event.
             *
             * @private
             * @param {Component} instance Target instance
             */
            unsubscribeAll: function (instance) {
                $listeners.forEach(function (listener, event) {
                    instance.off(event, listener);
                });
                $listeners.clear();
            },
            /**
             * Logs formatted data.
             *
             * @private
             * @param {string} msg Log message
             * @param {*} next The current value
             * @param {*} [prev] An optional previous value
             * @param {*[]} args Additional arguments
             */
            log: function (msg, next, prev, args) {
                if (process.env.NODE_ENV !== 'production') {
                    var head = [
                        "%c" + label + ": %c" + msg,
                        'color: gray; font-weight: lighter',
                        'color: green; font-weight: bolder',
                    ];
                    if (args && args.length > 0) {
                        head.push.apply(head, __spread(args));
                    }
                    console.groupCollapsed.apply(console, __spread(head));
                    if (prev !== undefined) {
                        console.log('%cprev state', 'color: gray; font-weight: bolder', prev);
                        console.log('%cnext state', 'color: green; font-weight: bolder', next);
                    }
                    else {
                        console.log('%cinitial state', 'color: gray; font-weight: bolder', next);
                    }
                    console.groupEnd();
                }
            },
        };
    };
    /* eslint-disable no-param-reassign */
    devtools = function (ComponentClass) {
        /**
         * Methods and properties added by devtools
         *
         * @name Component
         */
        var proto = ComponentClass.prototype;
        var mount = proto.mount, init = proto.init, destroy = proto.destroy;
        // function refTree(root: Component, tree: IObject) {
        //   if (root.$refsStore.size > 0) {
        //     root.$refsStore.forEach((ref, name) => {
        //       tree[name] = Object.create(null);
        //       refTree(ref, tree[name]);
        //     });
        //   }
        //   Object.defineProperty(tree, '$self', {
        //     enumerable: false,
        //     value: Object.assign(Object.create(null), {
        //       $raw: root,
        //       state: Object.assign(Object.create(null), root.state),
        //     }),
        //   });
        //   return tree;
        // }
        proto.mount = function mountDev() {
            var args = [];
            for (var _i = 0; _i < arguments.length; _i++) {
                args[_i] = arguments[_i];
            }
            mount.call.apply(mount, __spread([this], args));
            /**
             * A reference to a Component instance attached to its root element.
             *
             * @name $el.$yuzu
             * @type Component
             * @memberof Component
             */
            Object.defineProperty(this.$el, '$yuzu', {
                value: this,
                configurable: true,
            });
            // Object.defineProperty(this, '$$getTree', {
            //   enumerable: false,
            //   writable: false,
            //   value() {
            //     const tree = Object.create(null);
            //     return refTree(this, tree);
            //   },
            // });
            return this;
        };
        proto.destroy = function destroyDev() {
            var $el = this.$el;
            if ($el && $el.$yuzu) {
                delete $el.$yuzu;
            }
            delete this.$$logStart;
            delete this.$$logEnd;
            return destroy.call(this);
        };
        proto.init = function initDev(state) {
            Object.defineProperties(this, {
                /**
                 * ```js
                 * $$logStart([label], [event])
                 * ```
                 *
                 * Will initialize an event logger with a custom label. By default will automatically log any change to the instance `state` property.
                 *
                 * ?> Note that event loggers will be not available until the instance is initialized
                 *
                 * @memberof Component
                 * @param {string} [label] Log label. If not defined will fallback to: A `debugLabel` option, a static `displayName` property defined on the constructor, the component name
                 * @param {string|boolean} [listen="change:*"] Event to listen for or `false`.
                 */
                $$logStart: {
                    enumerable: false,
                    configurable: true,
                    value: function (label, listen) {
                        if (label === void 0) { label = null; }
                        if (listen === void 0) { listen = 'change:*'; }
                        var name = label ||
                            this.options.debugLabel ||
                            (this.constructor.displayName ||
                                this.constructor.name ||
                                'Component') + "#" + this.$uid;
                        if (!this.$$logger) {
                            this.$$logger = createStateLogger_1(name);
                        }
                        if (listen) {
                            this.$$logger.subscribe(this, listen);
                        }
                    },
                },
                /**
                 * ```js
                 * $$logEnd([event])
                 * ```
                 *
                 * Will stop an event logger either for a specific event or for every event.
                 *
                 * @memberof Component
                 * @param {string} [event] Event to unsubscribe for. If not provided the logger will be completely stopped and removed
                 */
                $$logEnd: {
                    enumerable: false,
                    configurable: true,
                    value: function (event) {
                        if (this.$$logger) {
                            if (event) {
                                this.$$logger.unsubscribe(this, event);
                                return;
                            }
                            this.$$logger.unsubscribeAll(this);
                            this.$$logger = undefined;
                        }
                    },
                },
            });
            return init.call(this, state);
        };
        // @ts-ignore: Devtools Hooks
        if (window.__YUZU_DEVTOOLS_GLOBAL_HOOK__) {
            // @ts-ignore
            window.__YUZU_DEVTOOLS_GLOBAL_HOOK__.init(Component);
        }
    };
    /* eslint-enable no-param-reassign */
}
else {
    devtools = function () { return undefined; };
}

export { Children, Component, DetachedComponent, devtools, extend, mount };
//# sourceMappingURL=index.m.js.map
